from collections.abc import Iterator
from contextlib import AbstractContextManager, contextmanager, nullcontext
from crystal import __version__ as crystal_version
from crystal import APP_NAME
from crystal.browser.entitytree import (
    EntityTree, ResourceGroupNode, RootResourceNode,
)
from crystal.browser.icons import TREE_NODE_ICONS
from crystal.browser.new_group import NewGroupDialog
from crystal.browser.new_root_url import ChangePrefixCommand, NewRootUrlDialog
from crystal.browser.preferences import PreferencesDialog
from crystal.browser.tasktree import TaskTree
from crystal.model import (
    Project, Resource, ResourceGroup, ResourceGroupSource, RootResource,
)
from crystal.progress import (
    CancelLoadUrls, DummyOpenProjectProgressListener,
    OpenProjectProgressListener,
)
from crystal.server import ProjectServer
from crystal.task import DownloadResourceGroupMembersTask, RootTask
from crystal.ui.actions import Action
from crystal.ui.BetterMessageDialog import BetterMessageDialog
from crystal.ui.log_drawer import LogDrawer
from crystal.ui.tree import DEFAULT_FOLDER_ICON_SET
from crystal.util.bulkheads import (
    capture_crashes_to, capture_crashes_to_stderr,
)
from crystal.util.ellipsis import EllipsisType
from crystal.util.finderinfo import get_hide_file_extension
from crystal.util.test_mode import tests_are_running
from crystal.util.unicode_labels import decorate_label
from crystal.util.url_prefix import (
    get_url_directory_prefix_for, get_url_domain_prefix_for,
)
from crystal.util.wx_bind import bind
from crystal.util.wx_dialog import (
    position_dialog_initially, set_dialog_or_frame_icon_if_appropriate,
    ShowModal,
)
from crystal.util.wx_timer import Timer, TimerError
from crystal.util.xos import (
    is_kde_or_non_gnome, is_linux, is_mac_os, is_windows, mac_version,
)
from crystal.util.xsqlite3 import (
    is_database_closed_error, is_database_gone_error,
)
from crystal.util.xthreading import (
    bg_call_later, fg_affinity, fg_call_and_wait, fg_call_later,
    set_is_quitting,
)
from functools import partial
import os
import sqlite3
import time
from typing import Optional
import webbrowser
import wx

_WINDOW_INNER_PADDING = 10


class MainWindow:
    _AUTOHIBERNATE_PERIOD = 1000 * 60 * 5  # 5 min, in milliseconds
    
    project: Project
    _frame: wx.Frame
    entity_tree: EntityTree
    task_tree: TaskTree
    
    # NOTE: Only changed when tests are running
    _last_created: 'Optional[MainWindow]'=None
    
    def __init__(self,
            project: Project,
            progress_listener: OpenProjectProgressListener | None=None,
            ) -> None:
        """
        Raises:
        * CancelOpenProject
        """
        if progress_listener is None:
            progress_listener = DummyOpenProjectProgressListener()
        
        self.project = project
        self._log_drawer = None  # type: Optional[LogDrawer]
        self._project_server = None  # type: Optional[ProjectServer]
        
        self._create_actions()
        
        frame_title: str
        filename_with_ext = os.path.basename(project.path)
        (filename_without_ext, filename_ext) = os.path.splitext(filename_with_ext)
        if is_windows() or is_kde_or_non_gnome():
            frame_title = f'{filename_without_ext} - {APP_NAME}'
        else:  # is_mac_os(); other
            extension_visible = (
                not get_hide_file_extension(project.path) if is_mac_os()
                else True
            )
            if extension_visible:
                frame_title = filename_with_ext
            else:
                frame_title = filename_without_ext
        
        # TODO: Rename: raw_frame -> frame,
        #               frame -> frame_content
        raw_frame = wx.Frame(None, title=frame_title, name='cr-main-window')
        try:
            # macOS: Define proxy icon beside the filename in the titlebar
            raw_frame.SetRepresentedFilename(project.path)
            # Define frame icon, if appropriate
            set_dialog_or_frame_icon_if_appropriate(raw_frame)
            
            # 1. Define *single* child with full content of the wx.Frame,
            #    so that LogDrawer can be created for this window later
            # 2. Add all controls to a root wx.Panel rather than to the
            #    raw wx.Frame directly so that tab traversal between child
            #    components works correctly.
            frame = wx.Panel(raw_frame)
            
            frame_sizer = wx.BoxSizer(wx.VERTICAL)
            frame.SetSizer(frame_sizer)
            
            splitter = wx.SplitterWindow(frame, style=wx.SP_3D|wx.SP_NO_XP_THEME|wx.SP_LIVE_UPDATE, name='cr-main-window-splitter')
            splitter.SetSashGravity(1.0)
            splitter.SetMinimumPaneSize(20)
            
            entity_pane = self._create_entity_pane(splitter, progress_listener)
            task_pane = self._create_task_pane(splitter)
            splitter.SplitHorizontally(entity_pane, task_pane, -task_pane.GetBestSize().height)
            
            frame_sizer.Add(splitter, proportion=1, flag=wx.EXPAND)
            
            frame_sizer.Add(
                self._create_status_bar(frame),
                proportion=0,
                flag=wx.EXPAND)
            
            raw_frame.MenuBar = self._create_menu_bar(raw_frame)
            
            bind(raw_frame, wx.EVT_CLOSE, self._on_close_frame)
            
            frame.Fit()
            raw_frame.Fit()  # NOTE: Must Fit() before Show() here so that wxGTK actually fits correctly
            raw_frame.Show(True)
            
            # Define minimum size for main window
            min_width = entity_pane.GetBestSize().Width
            min_height = task_pane.GetBestSize().Height * 2
            raw_frame.MinSize = wx.Size(min_width, min_height)
            
            self._frame = raw_frame
        except:
            raw_frame.Destroy()
            raise
        
        self._unhibernate()
        
        # Auto-hibernate every few minutes, in case Crystal crashes
        self._autohibernate_timer = None
        try:
            self._autohibernate_timer = Timer(self._hibernate, self._AUTOHIBERNATE_PERIOD)
        except TimerError:
            pass
        
        # Export reference to self, if running tests
        if tests_are_running():
            MainWindow._last_created = self
    
    @property
    def _readonly(self) -> bool:
        return self.project.readonly
    
    # === Actions ===
    
    def _create_actions(self) -> None:
        # File
        self._new_project_action = Action(
            wx.ID_NEW,
            '&New Project...',
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('N')),
            # TODO: Support multiple open projects
            action_func=None,
            enabled=False)
        self._open_project_action = Action(
            wx.ID_OPEN,
            '&Open Project...',
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('O')),
            # TODO: Support multiple open projects
            action_func=None,
            enabled=False)
        self._close_project_action = Action(
            wx.ID_CLOSE,
            '',
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('W')),
            action_func=self._on_close_project,
            enabled=True)
        self._save_project_action = Action(
            wx.ID_SAVE,
            '',
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('S')),
            # TODO: Support untitled projects (that require an initial save)
            action_func=None,
            enabled=False)
        self._quit_action = Action(
            wx.ID_EXIT,
            '',
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('Q')),
            # NOTE: Action is bound to self._on_quit later manually
            action_func=None,
            enabled=True)
        
        # Edit
        if is_mac_os():
            mv = mac_version()
            if mv is not None and mv >= [13]:
                preferences_label = 'Settings...'
            else:
                preferences_label = 'Preferences...'
        else:
            preferences_label = ''  # OS default
        self._preferences_action = Action(
            wx.ID_PREFERENCES,
            preferences_label,
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord(',')),
            # NOTE: Action is bound to self._on_preferences later manually
            action_func=None,
            enabled=True,
            button_label=decorate_label('⚙️', '&Preferences...', ''))
        
        # Entity
        self._new_root_url_action = Action(wx.ID_ANY,
            'New &Root URL...',
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('R')),
            self._on_new_root_url,
            enabled=(not self._readonly),
            button_bitmap=TREE_NODE_ICONS()['entitytree_root_resource'],
            button_label='New &Root URL...')
        self._new_group_action = Action(wx.ID_ANY,
            'New &Group...',
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('G')),
            self._on_new_group,
            enabled=(not self._readonly),
            button_bitmap=dict(DEFAULT_FOLDER_ICON_SET())[wx.TreeItemIcon_Normal],
            button_label='New &Group...')
        self._edit_action = Action(wx.ID_ANY,
            '&Edit...',
            accel=wx.AcceleratorEntry(wx.ACCEL_NORMAL, wx.WXK_RETURN),
            action_func=self._on_edit_entity,
            enabled=False,
            button_label=decorate_label('✏️', '&Edit...', ''))
        self._forget_action = Action(wx.ID_ANY,
            '&Forget',
            wx.AcceleratorEntry(wx.ACCEL_CTRL, wx.WXK_BACK),
            self._on_forget_entity,
            enabled=False,
            button_label=decorate_label('✖️', '&Forget', ''))
        self._download_action = Action(wx.ID_ANY,
            '&Download',
            wx.AcceleratorEntry(wx.ACCEL_CTRL, wx.WXK_RETURN),
            self._on_download_entity,
            enabled=False,
            button_label=decorate_label('⬇', '&Download', ''))
        self._update_members_action = Action(wx.ID_ANY,
            'Update &Members',
            accel=None,
            action_func=self._on_update_group_members,
            enabled=False,
            button_label=decorate_label('🔎', 'Update &Members', ' '))
        self._view_action = Action(wx.ID_ANY,
            '&View',
            # TODO: Consider adding Space as a alternate accelerator
            #wx.AcceleratorEntry(wx.ACCEL_NORMAL, wx.WXK_SPACE),
            wx.AcceleratorEntry(wx.ACCEL_CTRL|wx.ACCEL_SHIFT, ord('O')),
            self._on_view_entity,
            enabled=False,
            button_label=decorate_label('👀', '&View', ' '))
        
        # HACK: Gather all actions indirectly by inspecting fields
        self._actions = [a for a in self.__dict__.values() if isinstance(a, Action)]
    
    # === Menubar ===
    
    def _create_menu_bar(self, raw_frame: wx.Frame) -> wx.MenuBar:
        file_menu = wx.Menu()
        self._new_project_action.append_menuitem_to(file_menu)
        self._open_project_action.append_menuitem_to(file_menu)
        file_menu.AppendSeparator()
        self._close_project_action.append_menuitem_to(file_menu)
        self._save_project_action.append_menuitem_to(file_menu)
        # Append Quit menuitem
        if True:
            self._quit_action.append_menuitem_to(file_menu)
            # NOTE: Can only intercept wx.EVT_MENU for wx.ID_EXIT on an wx.Frame
            #       on macOS. In particular cannot intercept on the File wx.Menu.
            bind(raw_frame, wx.EVT_MENU, self._on_quit)
        
        edit_menu = wx.Menu()
        edit_menu.Append(wx.ID_UNDO, '').Enabled = False
        edit_menu.Append(wx.ID_REDO, '').Enabled = False
        edit_menu.AppendSeparator()
        edit_menu.Append(wx.ID_CUT, '').Enabled = False
        edit_menu.Append(wx.ID_COPY, '').Enabled = False
        edit_menu.Append(wx.ID_PASTE, '').Enabled = False
        # Append Preferences menuitem
        if True:
            if not is_mac_os():
                edit_menu.AppendSeparator()
            # NOTE: On macOS the Preferences menuitem will actually be positioned
            #       in the [Application Name] menu rather than the Edit menu.
            self._preferences_action.append_menuitem_to(edit_menu)
            # NOTE: Can only intercept wx.EVT_MENU for wx.ID_PREFERENCES on an wx.Frame
            #       on macOS. In particular cannot intercept on the Edit wx.Menu.
            bind(raw_frame, wx.EVT_MENU, self._on_preferences)
        
        self._entity_menu = entity_menu = wx.Menu()
        self._new_root_url_action.append_menuitem_to(entity_menu)
        self._new_group_action.append_menuitem_to(entity_menu)
        self._edit_action.append_menuitem_to(entity_menu)
        self._forget_action.append_menuitem_to(entity_menu)
        entity_menu.AppendSeparator()
        self._download_action.append_menuitem_to(entity_menu)
        self._update_members_action.append_menuitem_to(entity_menu)
        self._view_action.append_menuitem_to(entity_menu)
        entity_menu.AppendSeparator()
        if True:
            (cup_mis, on_attach_menuitems) = \
                self.entity_tree.create_change_url_prefix_menuitems_for(
                    node=None, menu_type='top_level')
            for mi in cup_mis:
                entity_menu.Append(mi)
            on_attach_menuitems()
            self._change_url_prefix_menuitems = cup_mis
            
            bind(entity_menu, wx.EVT_MENU_OPEN, self._on_entity_menu_open)
            bind(entity_menu, wx.EVT_MENU, self._on_change_url_prefix_menuitem_selected)
        
        menubar = wx.MenuBar()
        menubar.Append(file_menu, 'File')
        menubar.Append(edit_menu, 'Edit')
        menubar.Append(entity_menu, 'Entity')
        return menubar
    
    def _on_entity_menu_open(self, event: wx.MenuEvent) -> None:
        menu = self._entity_menu  # cache
        
        # Locate old _change_url_prefix_menuitems
        assert len(self._change_url_prefix_menuitems) >= 1
        first_old_mi = self._change_url_prefix_menuitems[0]
        (first_old_mi2, first_old_mi_offset) = \
            menu.FindChildItem(first_old_mi.Id)
        assert first_old_mi2 == first_old_mi
        
        # Remove/dispose old _change_url_prefix_menuitems
        for mi in reversed(self._change_url_prefix_menuitems):
            menu.Remove(mi.Id)
        self._change_url_prefix_menuitems.clear()
        
        # Create new _change_url_prefix_menuitems
        (cup_mis, on_attach_menuitems) = \
            self.entity_tree.create_change_url_prefix_menuitems_for(
                node=self.entity_tree.selected_node,
                menu_type='top_level')
        for (mi_offset, mi) in enumerate(cup_mis, start=first_old_mi_offset):
            menu.Insert(mi_offset, mi)
        on_attach_menuitems()
        self._change_url_prefix_menuitems = cup_mis
    
    def _on_change_url_prefix_menuitem_selected(self, event: wx.MenuEvent) -> None:
        self.entity_tree.on_change_url_prefix_menuitem_selected(
            event, self.entity_tree.selected_node)
    
    # === Entity Pane: Init ===
    
    def _create_entity_pane(self, parent, progress_listener: OpenProjectProgressListener):
        """
        Raises:
        * CancelOpenProject
        """
        pane = wx.Panel(parent)
        pane_sizer = wx.BoxSizer(wx.VERTICAL)
        pane.SetSizer(pane_sizer)
        
        pane_sizer.Add(
            self._create_entity_pane_content(pane, progress_listener),
            proportion=1,
            flag=wx.EXPAND|wx.ALL,
            border=_WINDOW_INNER_PADDING)
        
        return pane
    
    def _create_entity_pane_content(self, parent: wx.Window, progress_listener: OpenProjectProgressListener):
        """
        Raises:
        * CancelOpenProject
        """
        content_sizer = wx.BoxSizer(wx.VERTICAL)
        content_sizer.Add(
            self._create_entity_tree(parent, progress_listener),
            proportion=1,
            flag=wx.EXPAND)
        content_sizer.AddSpacer(_WINDOW_INNER_PADDING)
        content_sizer.Add(self._create_button_bar(parent), flag=wx.EXPAND)
        return content_sizer
    
    def _create_entity_tree(self, parent: wx.Window, progress_listener: OpenProjectProgressListener):
        """
        Raises:
        * CancelOpenProject
        """
        self.entity_tree = EntityTree(parent, self.project, progress_listener)
        bind(self.entity_tree.peer, wx.EVT_TREE_SEL_CHANGED, self._on_selected_entity_changed)
        self._on_selected_entity_changed()
        
        return self.entity_tree.peer
    
    def _create_button_bar(self, parent: wx.Window):
        readonly = self._readonly  # cache
        
        new_root_url_button = self._new_root_url_action.create_button(parent, name='cr-add-url-button')
        new_group_button = self._new_group_action.create_button(parent, name='cr-add-group-button')
        edit_entity_button = self._edit_action.create_button(parent, name='cr-edit-button')
        forget_entity_button = self._forget_action.create_button(parent, name='cr-forget-button')
        download_button = self._download_action.create_button(parent, name='cr-download-button')
        update_members_button = self._update_members_action.create_button(
            parent, name='cr-update-members-button')
        
        view_button = self._view_action.create_button(parent, name='cr-view-button')
        
        content_sizer = wx.BoxSizer(wx.HORIZONTAL)
        content_sizer.Add(new_root_url_button)
        content_sizer.AddSpacer(_WINDOW_INNER_PADDING)
        content_sizer.Add(new_group_button)
        content_sizer.AddSpacer(_WINDOW_INNER_PADDING)
        content_sizer.Add(edit_entity_button)
        content_sizer.AddSpacer(_WINDOW_INNER_PADDING)
        content_sizer.Add(forget_entity_button)
        content_sizer.AddSpacer(_WINDOW_INNER_PADDING * 2)
        content_sizer.AddStretchSpacer()
        content_sizer.Add(download_button)
        content_sizer.AddSpacer(_WINDOW_INNER_PADDING)
        content_sizer.Add(update_members_button)
        content_sizer.AddSpacer(_WINDOW_INNER_PADDING)
        content_sizer.Add(view_button)
        return content_sizer
    
    # === Entity Pane: Properties ===
    
    @property
    def _suggested_url_or_url_pattern_for_selection(self) -> str | None:
        selected_entity = self.entity_tree.selected_entity
        
        if isinstance(selected_entity, (Resource, RootResource)):
            return selected_entity.resource.url
        elif isinstance(selected_entity, ResourceGroup):
            return selected_entity.url_pattern
        else:
            return self.project.default_url_prefix
    
    @property
    def _suggested_source_for_selection(self) -> ResourceGroupSource:
        return self.entity_tree.source_of_selection
    
    @property
    def _suggested_name_for_selection(self) -> str | None:
        return self.entity_tree.name_of_selection
    
    # === Operations ===
    
    def close(self) -> None:
        """
        Closes this window.
        
        The caller is still responsible for closing the underlying Project.
        """
        dummy_event = wx.CommandEvent()
        self._on_close_project(dummy_event)
    
    def __enter__(self) -> 'MainWindow':
        return self
    
    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.close()
    
    # === File Menu: Events ===
    
    def _on_close_project(self, event: wx.CommandEvent) -> None:
        self._frame.Close()  # will trigger call to _on_close_frame()
    
    def _on_quit(self, event: wx.CommandEvent) -> None:
        if event.Id == wx.ID_EXIT:
            set_is_quitting()
            self._frame.Close()  # will trigger call to _on_close_frame()
        else:
            event.Skip()
    
    @fg_affinity
    def _on_close_frame(self, event: wx.CloseEvent) -> None:
        """
        Closes this window, disposing any related resources.
        """
        if self._autohibernate_timer is not None:
            self._autohibernate_timer.stop()
        
        self._hibernate()
        
        # Dispose resources created in MainWindow.start_server(), in reverse order
        if self._project_server is not None:
            self._project_server.close()
        if self._log_drawer is not None:
            self._log_drawer.close()
        
        # Dispose resources created in MainWindow.__init__(), in reverse order
        if True:
            self.task_tree.dispose()
            self.entity_tree.dispose()
            
            # Dispose actions so that they don't try to interact with
            # wx.Objects in the frame while the frame is being deleted
            for a in self._actions:
                a.dispose()
        
        event.Skip()  # continue dispose of frame
    
    # === Entity Pane: New/Edit Root Url ===
    
    def _on_new_root_url(self, event: wx.CommandEvent) -> None:
        NewRootUrlDialog(
            self._frame,
            self._on_new_root_url_dialog_ok,
            url_exists_func=self._root_url_exists,
            initial_url=self._suggested_url_or_url_pattern_for_selection or '',
            initial_name=self._suggested_name_for_selection or '',
            initial_set_as_default_domain=(self.project.default_url_prefix is None),
            initial_set_as_default_directory=False,
        )
    
    def _root_url_exists(self, url: str) -> bool:
        r = self.project.get_resource(url)
        if r is None:
            return False
        rr = self.project.get_root_resource(r)
        return rr is not None
    
    @fg_affinity
    def _on_new_root_url_dialog_ok(self,
            name: str,
            url: str,
            change_prefix_command: ChangePrefixCommand,
            download_immediately: bool,
            create_group: bool,
            ) -> None:
        if url == '':
            raise ValueError('Invalid blank URL')
        
        try:
            rr = RootResource(self.project, name, Resource(self.project, url))
        except RootResource.AlreadyExists:
            raise ValueError('Invalid duplicate URL')
        
        if isinstance(change_prefix_command, EllipsisType):
            pass
        elif change_prefix_command is None:
            self.entity_tree.clear_default_url_prefix()
        else:
            self.entity_tree.set_default_url_prefix(*change_prefix_command)
        
        if create_group:
            assert url.endswith('/')
            rg = ResourceGroup(self.project, '', url + '**', source=rr)
            
            if download_immediately:
                rg.download(needs_result=False)
        else:
            if download_immediately:
                rr.download(needs_result=False)
    
    @fg_affinity
    def _on_edit_root_url_dialog_ok(self,
            rr: RootResource,
            name: str,
            url: str,
            change_prefix_command: ChangePrefixCommand,
            download_immediately: bool,
            create_group: bool,
            ) -> None:
        if url != rr.url:
            raise ValueError()
        
        if name != rr.name:
            rr.name = name
            
            # TODO: This update should happen in response to an event
            #       fired by the entity itself.
            self.entity_tree.root.update_title_of_descendants()  # update names in titles
        
        if isinstance(change_prefix_command, EllipsisType):
            pass
        elif change_prefix_command is None:
            self.entity_tree.clear_default_url_prefix()
        else:
            self.entity_tree.set_default_url_prefix(*change_prefix_command)
        
        assert download_immediately == False
        assert create_group == False
    
    # === Entity Pane: New/Edit Group ===
    
    def _on_new_group(self, event: wx.CommandEvent) -> None:
        try:
            NewGroupDialog(
                self._frame,
                self._on_new_group_dialog_ok,
                self.project,
                saving_source_would_create_cycle_func=lambda source: False,
                initial_url_pattern=self._suggested_url_or_url_pattern_for_selection or '',
                initial_source=self._suggested_source_for_selection,
                initial_name=self._suggested_name_for_selection or '')
        except CancelLoadUrls:
            pass
    
    @fg_affinity
    def _on_new_group_dialog_ok(self,
            name: str,
            url_pattern: str,
            source: ResourceGroupSource,
            do_not_download: bool,
            download_immediately: bool,
            ) -> None:
        # TODO: Validate user input:
        #       * Is url_pattern empty?
        #       * Is url_pattern already taken?
        rg = ResourceGroup(
            self.project, name, url_pattern, source,
            do_not_download=do_not_download)
        
        if download_immediately:
            rg.download(needs_result=False)
    
    @fg_affinity
    def _on_edit_group_dialog_ok(self,
            rg: ResourceGroup,
            name: str,
            url_pattern: str,
            source: ResourceGroupSource,
            do_not_download: bool,
            download_immediately: bool,
            ) -> None:
        if url_pattern != rg.url_pattern:
            raise ValueError()
        (rg.name, rg.source, rg.do_not_download) = (name, source, do_not_download)
        
        # TODO: This update should happen in response to an event
        #       fired by the entity itself.
        self.entity_tree.root.update_title_of_descendants()  # update names in titles
        
        assert download_immediately == False
    
    def _saving_source_would_create_cycle(self, rg: ResourceGroup, source: ResourceGroupSource) -> bool:
        ancestor_source = source  # type: ResourceGroupSource
        while ancestor_source is not None:
            if ancestor_source == rg:
                return True
            if isinstance(ancestor_source, ResourceGroup):
                ancestor_source = ancestor_source.source  # reinterpret
            else:
                ancestor_source = None
        return False
    
    # === Entity Pane: Other Commands ===
    
    def _on_edit_entity(self, event) -> None:
        selected_entity = self.entity_tree.selected_entity
        assert selected_entity is not None
        
        if isinstance(selected_entity, RootResource):
            rr = selected_entity
            selection_urllike = rr.resource.url
            selection_domain_prefix = get_url_domain_prefix_for(selection_urllike)
            selection_dir_prefix = get_url_directory_prefix_for(selection_urllike)
            NewRootUrlDialog(
                self._frame,
                partial(self._on_edit_root_url_dialog_ok, rr),
                url_exists_func=self._root_url_exists,
                initial_url=rr.url,
                initial_name=rr.name,
                initial_set_as_default_domain=(
                    selection_domain_prefix is not None and
                    self.project.default_url_prefix == selection_domain_prefix
                ),
                initial_set_as_default_directory=(
                    selection_dir_prefix is not None and
                    selection_dir_prefix != selection_domain_prefix and
                    self.project.default_url_prefix == selection_dir_prefix
                ),
                allow_set_as_default_domain_or_directory=(
                    selection_domain_prefix is not None or
                    selection_dir_prefix is not None
                ),
                is_edit=True,
            )
        elif isinstance(selected_entity, ResourceGroup):
            rg = selected_entity
            try:
                NewGroupDialog(
                    self._frame,
                    partial(self._on_edit_group_dialog_ok, rg),
                    self.project,
                    saving_source_would_create_cycle_func=
                        partial(self._saving_source_would_create_cycle, rg),
                    initial_url_pattern=rg.url_pattern,
                    initial_source=rg.source,
                    initial_name=rg.name,
                    initial_do_not_download=rg.do_not_download,
                    is_edit=True)
            except CancelLoadUrls:
                pass
        else:
            raise AssertionError()
    
    def _on_forget_entity(self, event) -> None:
        selected_entity = self.entity_tree.selected_entity
        assert selected_entity is not None
        
        selected_entity.delete()
    
    def _on_download_entity(self, event) -> None:
        selected_entity = self.entity_tree.selected_entity
        assert selected_entity is not None
        
        # Show progress dialog in advance if will need to load all project URLs
        if isinstance(selected_entity, ResourceGroup):
            try:
                selected_entity.project.load_urls()
            except CancelLoadUrls:
                return
        
        # Show progress dialog if it will likely take a long time to start the download
        if DownloadResourceGroupMembersTask._LAZY_LOAD_CHILDREN:
            progress_dialog = None  # type: Optional[wx.ProgressDialog]
        else:
            if (isinstance(selected_entity, ResourceGroup) and
                    len(selected_entity.members) >= 2000):  # TODO: Tune threshold
                progress_dialog = wx.ProgressDialog(
                    title='Starting download',
                    message=f'Starting download of {len(selected_entity.members):n} members...',
                    parent=self._frame,
                    style=wx.PD_ELAPSED_TIME)
                progress_dialog.Name = 'cr-starting-download'
                progress_dialog.Pulse(progress_dialog.Message)  # make progress bar indeterminate
                progress_dialog.Show()
                
                # Update the elapsed time every 1 second
                # 
                # NOTE: It would be simpler to implement this logic with wx.Timer
                #       but wx.Timer seems to not work well if very many lambdas
                #       are scheduled with fg_call_later at the same time.
                @capture_crashes_to_stderr
                def elapsed_time_updater() -> None:
                    while True:
                        time.sleep(1.0)
                        def fg_task() -> bool:
                            if progress_dialog is not None:
                                progress_dialog.Pulse(progress_dialog.Message)
                                return True
                            else:
                                return False
                        still_ticking = fg_call_and_wait(fg_task)
                        if not still_ticking:
                            break
                bg_call_later(elapsed_time_updater)
            else:
                progress_dialog = None
        
        # Run download() on a background thread because it can take a long time
        # to instantiate the tree of related download tasks (when _LAZY_LOAD_CHILDREN == False)
        # 
        # NOTE: Loudly crashes the entire scheduler thread upon failure.
        #       If this failure mode ends up happening commonly,
        #       suggest implementing a less drastic failure mode.
        @capture_crashes_to(self.project.root_task)
        def bg_task() -> None:
            assert selected_entity is not None
            
            # Start download
            selected_entity.download(needs_result=False)
            
            # Close progress dialog, if applicable
            if progress_dialog is not None:
                def fg_task() -> None:
                    nonlocal progress_dialog
                    assert progress_dialog is not None
                    progress_dialog.Destroy()
                    progress_dialog = None  # unexport
                fg_call_and_wait(fg_task)
        bg_call_later(bg_task)
    
    def _on_update_group_members(self, event):
        selected_entity = self.entity_tree.selected_entity
        selected_entity.update_members()
    
    def _on_view_entity(self, event) -> None:
        # TODO: If the server couldn't be started (ex: due to the default port being in
        #       use), report an appropriate error.
        project_server = self.start_server()
        
        selected_entity = self.entity_tree.selected_entity
        assert isinstance(selected_entity, (Resource, RootResource))
        archive_url = selected_entity.resource.url
        request_url = project_server.get_request_url(archive_url)
        
        def open_browser_to_url() -> None:
            if is_linux():
                # HACK: Firefox on Linux starts with an error if
                #       the current working directory is not writable,
                #       so make sure the current working directory is
                #       writable before potentially starting Firefox
                open_browser_context = self._cwd_set_to_writable_dir()  # type: AbstractContextManager
            else:
                open_browser_context = nullcontext()
            with open_browser_context:
                # NOTE: Can block for as long as 3 seconds on Linux
                webbrowser.open(request_url)
        
        # HACK: If LogDrawer is using _WindowPairingStrategy.FLOAT_AND_RAISE_ON_ACTIVATE,
        #       make sure the log drawer is done bringing itself and MainWindow to front
        #       before bringing a browser to the front on top of them
        if os.environ.get('CRYSTAL_RUNNING_TESTS', 'False') == 'False':
            wx.CallLater(10, open_browser_to_url)
        else:
            # NOTE: During tests it's easier to spy on this call if it's NOT
            #       deferred inside wx.CallLater
            open_browser_to_url()
    
    @contextmanager
    def _cwd_set_to_writable_dir(self) -> Iterator[None]:
        assert is_linux(), 'This function only supports Linux'
        new_cwd = os.environ.get('HOME', '/')
        old_cwd = os.getcwd()  # capture
        os.chdir(new_cwd)
        try:
            yield
        finally:
            os.chdir(old_cwd)
    
    def start_server(self) -> 'ProjectServer':
        """
        Starts an HTTP server that serves pages from this project.
        
        If an HTTP server is already running, does nothing.
        """
        if self._project_server is None:
            self._log_drawer = LogDrawer(parent=self._frame)
            self._project_server = ProjectServer(self.project, stdout=self._log_drawer.writer)
        
        return self._project_server
    
    # === Entity Pane: Events ===
    
    def _on_selected_entity_changed(self, event: wx.TreeEvent | None=None) -> None:
        selected_entity = self.entity_tree.selected_entity  # cache
        
        readonly = self._readonly  # cache
        self._edit_action.enabled = (
            (not readonly) and
            isinstance(selected_entity, (ResourceGroup, RootResource)))
        self._forget_action.enabled = (
            (not readonly) and
            isinstance(selected_entity, (ResourceGroup, RootResource)))
        self._download_action.enabled = (
            (not readonly) and
            selected_entity is not None)
        self._update_members_action.enabled = (
            (not readonly) and
            isinstance(selected_entity, ResourceGroup))
        self._view_action.enabled = (
            isinstance(selected_entity, (Resource, RootResource)))
    
    # === Task Pane: Init ===
    
    def _create_task_pane(self, parent: wx.Window) -> wx.Window:
        pane = wx.Panel(parent)
        pane_sizer = wx.BoxSizer(wx.VERTICAL)
        pane.SetSizer(pane_sizer)
        
        pane_sizer.Add(
            self._create_task_pane_content(pane), 
            proportion=1, 
            flag=wx.EXPAND|wx.ALL, 
            border=_WINDOW_INNER_PADDING)
        
        return pane
    
    def _create_task_pane_content(self, parent: wx.Window) -> wx.Window:
        content_sizer = wx.BoxSizer(wx.VERTICAL)
        content_sizer.Add(self._create_task_tree(parent), proportion=1, flag=wx.EXPAND)
        return content_sizer
    
    def _create_task_tree(self, parent: wx.Window) -> wx.Window:
        self.task_tree = TaskTree(parent, self.project.root_task)
        if is_mac_os():
            self.task_tree.peer.SetBackgroundColour(
                wx.Colour(254, 254, 254))  # pure white
        
        return self.task_tree.peer
    
    # === Task Pane: Other Commands ===
    
    @capture_crashes_to_stderr
    def _hibernate(self) -> None:
        if self.project.readonly:
            return
        
        try:
            self.project.hibernate_tasks()
        except (sqlite3.DatabaseError, sqlite3.ProgrammingError) as e:
            # Ignore certain types of I/O errors while closing
            io_error = (
                # Disk containing the database may have unmounted expectedly
                is_database_gone_error(e) or
                # Automated tests are simulating unmount of the disk
                # containing the database, using _close_project_abruptly()
                (tests_are_running() and is_database_closed_error(e))
            )
            if not io_error:
                raise
    
    @capture_crashes_to_stderr
    def _unhibernate(self) -> None:
        if self.project.readonly:
            return
        
        def confirm_unhibernate_tasks() -> bool:
            dialog = wx.MessageDialog(
                self._frame,
                message=(
                    'Downloads were running when this project was last closed. '
                    'Resume them?'
                ),
                caption='Resume Downloads?',
                style=wx.OK|wx.CANCEL,
            )
            dialog.Name = 'cr-resume-downloads'
            with dialog:
                dialog.SetOKCancelLabels('Resume', wx.ID_CANCEL)
                dialog.SetEscapeId(wx.ID_CANCEL)
                position_dialog_initially(dialog)
                choice = ShowModal(dialog)
            should_resume = (choice == wx.ID_OK)
            return should_resume
        self.project.unhibernate_tasks(confirm_unhibernate_tasks)
    
    # === Status Bar: Init ===
    
    def _create_status_bar(self, parent: wx.Window) -> wx.Sizer:
        readonly = self._readonly  # cache
        
        pane = parent
        pane_sizer = wx.BoxSizer(wx.HORIZONTAL)
        
        version_label = wx.StaticText(pane, label=f'Crystal v{crystal_version}')
        
        preferences_button = self._preferences_action.create_button(pane, name='cr-preferences-button')
        bind(preferences_button, wx.EVT_BUTTON, self._on_preferences)
        
        if readonly:
            rwi_label = '🔒' if not is_windows() else 'Read only'
            rwi_tooltip = 'Read only project'  # type: str
        else:
            rwi_label = '✏️' if not is_windows() else 'Writable'
            rwi_tooltip = 'Writable project'
        read_write_icon = wx.StaticText(pane, label=rwi_label, name='cr-read-write-icon')
        read_write_icon.SetToolTip(rwi_tooltip)
        
        pane_sizer.Add(
            version_label,
            proportion=1,
            flag=wx.CENTER|wx.EXPAND|wx.ALL,
            border=_WINDOW_INNER_PADDING)
        pane_sizer.Add(
            preferences_button,
            flag=wx.CENTER|wx.ALL &~ wx.RIGHT,
            border=_WINDOW_INNER_PADDING)
        pane_sizer.Add(
            read_write_icon,
            flag=wx.CENTER|wx.ALL,
            border=_WINDOW_INNER_PADDING)
        
        return pane_sizer
    
    # === Status Bar: Events ===
    
    def _on_preferences(self, event: wx.MenuEvent | wx.CommandEvent) -> None:
        if event.Id == wx.ID_PREFERENCES or isinstance(event.EventObject, wx.Button):
            PreferencesDialog(self._frame, self.project)
        else:
            event.Skip()



from sortedcontainers import SortedList
from typing_extensions import override


class BlackHoleSortedList(SortedList):
    """
    A SortedList that is always empty. Silently ignores requests to add items.
    """
    def __init__(self, iterable=None, key=None):
        super().__init__()
    
    @override
    def add(self, value):
        pass
    
    @override
    def update(self, iterable):
        pass
    
    @override
    def __iadd__(self, other):
        return self
    
    @override
    def __imul__(self, num):
        return self


BLACK_HOLE_SORTED_LIST = BlackHoleSortedList()



"""
Implements an HTTP server that serves resource revisions from a Project.
Runs on its own daemon thread.
"""

from __future__ import annotations

from collections.abc import Callable, Generator
from crystal.doc.generic import Document, Link
from crystal.doc.html.soup import HtmlDocument
from crystal.model import (
    Project, Resource, ResourceGroup, ResourceRevision, RootResource,
)
from crystal.util.bulkheads import capture_crashes_to_stderr
from crystal.util.cli import (
    print_error, print_info, print_success, print_warning,
)
from crystal.util.ports import is_port_in_use, is_port_in_use_error
from crystal.util.test_mode import tests_are_running
from crystal.util.xthreading import (
    bg_affinity, bg_call_later, fg_call_and_wait,
    run_thread_switching_coroutine, SwitchToThread,
)
import datetime
from html import escape as html_escape  # type: ignore[attr-defined]
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, HTTPServer
from io import TextIOBase
import re
from textwrap import dedent
from typing import Literal, Optional
from typing_extensions import override
from urllib.parse import parse_qs, urljoin, urlparse, urlunparse

_DEFAULT_SERVER_PORT = 2797  # CRYS on telephone keypad


Verbosity = Literal['normal', 'indent']


class ProjectServer:
    """Runs the archive server on a thread."""
    
    # NOTE: Only changed when tests are running
    _last_created: Optional[ProjectServer]=None
    
    def __init__(self,
            project: Project,
            port: int | None=None,
            *, verbosity: Verbosity='normal',
            stdout: TextIOBase | None=None,
            ) -> None:
        if port is None:
            port = _DEFAULT_SERVER_PORT
            try_other_ports = True
        else:
            try_other_ports = False
        
        self._project = project
        self._verbosity = verbosity
        self._stdout = stdout
        
        # Start server on port, looking for alternative open port if needed
        while True:
            # NOTE: Must explicitly check whether port in use because Windows
            #       appears not to raise an exception when opening a _HttpServer
            #       on a port that is in use, unlike macOS and Linux
            if try_other_ports and is_port_in_use(port):
                pass
            else:
                try:
                    address = ('', port)
                    server = _HttpServer(address, _RequestHandler)
                except Exception as e:
                    if try_other_ports and is_port_in_use_error(e):
                        pass
                    else:
                        raise
                else:
                    break
            
            # Try another port
            port += 1
            continue
        server.project = project
        server.verbosity = verbosity
        server.stdout = stdout
        
        self._server = server
        self._port = port
        
        # NOTE: It's very hard to crash _HttpServer itself such that an
        #       unhandled exception would escape to this level.
        #       Most errors already get routed to _HttpServer.handle_error().
        #       For those errors that somehow manage to escape anyway,
        #       just route them to stderr for now.
        @capture_crashes_to_stderr
        def bg_task() -> None:
            try:
                if verbosity == 'normal':
                    print_success('Server started on port %s.' % port, file=self._stdout)
                self._server.serve_forever()
            finally:
                self._server.server_close()
        bg_call_later(bg_task, daemon=True)
        
        # Export reference to self, if running tests
        if tests_are_running():
            ProjectServer._last_created = self
    
    def close(self) -> None:
        self._server.shutdown()
    
    # === Properties ===
    
    @property
    def project(self) -> Project:
        return self._project
    
    @property
    def port(self) -> int:
        return self._port
    
    # === Utility ===
    
    def get_request_url(self, archive_url: str) -> str:
        """
        Given the absolute URL of a resource, returns the URL that should be used to
        request it from the project server.
        """
        return get_request_url(archive_url, self._port, self._project.default_url_prefix)


def get_request_url(
        archive_url: str,
        port: int | None=None,
        project_default_url_prefix: str | None=None,
        ) -> str:
    """
    Given the absolute URL of a resource, returns the URL that should be used to
    request it from the project server.
    """
    if port is None:
        port = _DEFAULT_SERVER_PORT
    
    request_host = 'localhost:%s' % port
    return _RequestHandler.get_request_url_with_host(
        archive_url, request_host, project_default_url_prefix)


# ------------------------------------------------------------------------------

_REQUEST_PATH_IN_ARCHIVE_RE = re.compile(r'^/_/([^/]+)/(.+)$')

_PIN_DATE_JS_PATH_RE = re.compile(r'^/_/crystal/pin_date\.js\?t=([0-9]+)$')
_PIN_DATE_JS_PATH_PREFIX = '/_/crystal/pin_date.js?t='

# Set of archived headers that may be played back as-is
_HEADER_ALLOWLIST = {
    'date',
    'vary',
    'content-disposition',
    'content-language',
    'content-type',
    'server',
    'last-modified',
    'etag',
    'content-encoding',
    'x-xss-protection',
    'x-content-type-options',
    'x-frame-options',
    'x-download-options',
    'x-permitted-cross-domain-policies',
    
    # CORS
    'access-control-allow-origin',
    'access-control-allow-methods',
    'access-control-expose-headers',
    
    # CORP
    'cross-origin-resource-policy',
    
    # COOP
    'cross-origin-opener-policy',
    
    # Timing
    'timing-allow-origin',  # enable cross-origin access to timing API
    'server-timing',
    
    # Infrastructure-specific: AWS Cloudfront
    'x-amz-cf-id',
    'x-amz-cf-pop',
    'x-amz-storage-class',
    
    # Infrastructure-specific: AWS S3
    'x-amz-id-2',
    'x-amz-request-id',
    'x-amz-version-id',
    
    # Infrastructure-specific: Cloudflare
    'cf-cache-status',
    'cf-ray',
    'cf-request-id',
    'cf-bgj',
    'cf-polished',
    
    # Infrastructure-specific: Google Cloud
    'x-goog-generation',
    'x-goog-hash',
    'x-goog-metageneration',
    'x-goog-storage-class',
    'x-goog-stored-content-encoding',
    'x-goog-stored-content-length',
    'x-guploader-uploadid',
    
    # Infrastructure-specific: Fastly
    'detected-user-agent',
    'fastly-restarts',
    'normalized-user-agent',
    'request_came_from_shield',
    'surrogate-control',
    'x-served-by',
    
    # Infrastructure-specific: Envoy
    'x-envoy-attempt-count',
    'x-envoy-upstream-service-time',
    
    # Service-specific: imgix
    'x-imgix-id',
    
    # Site-specific: Substack
    'x-cluster',
}
# Set of archived headers known to cause problems if blindly played back
_HEADER_DENYLIST = {
    # Connection
    'transfer-encoding',# overridden by this web server
    'content-length',   # overridden by this web server
    'accept-ranges',    # partial content ranges not supported by this server
    'connection',       # usually: keep-alive
    'keep-alive',
    'via',              # this web server is not a proxy
    
    # Cache
    'cache-control',
    'age',
    'expires',
    'pragma',           # usually: no-cache
    
    # Cookie
    'set-cookie',       # don't allow cookies to be set by archived site
    'p3p',              # never allow third party cookies
    
    # HTTPS & Certificates
    'expect-ct',        # don't require Certificate Transparency
    'strict-transport-security',  # don't require HTTPS
    
    # Links (uninteresting)
    # NOTE: Details at: https://github.com/davidfstr/Crystal-Web-Archiver/issues/75
    'link',
    
    # Logging
    'content-security-policy-report-only',
                        # don't enable CSP logging
    'cross-origin-embedder-policy-report-only',
                        # don't enable COEP logging
    'cross-origin-opener-policy-report-only',
                        # don't enable COOP logging
    'nel',              # don't enable network request logging
    'report-to',        # don't enable Content Security Policy logging
    
    # Referer
    'referrer-policy',  # don't omit referer because used to dynamically rewrite links
    
    # Protocol Upgrades
    'upgrade',          # we don't support HTTP/2, QUIC, or any alternate protocol advertised by this
    'alt-svc',          # we don't support HTTP/2, QUIC, or any alternate protocol advertised by this
    
    # X-RateLimit
    'x-ratelimit-limit',
    'x-ratelimit-remaining',
    'x-ratelimit-count',
    'x-ratelimit-reset',
    
    # Ignored non-problematic headers
    # TODO: Consider moving these headers to the _HEADER_ALLOWLIST
    'x-powered-by',
    'x-monk',
    'x-cache',
    'x-ecn-p',
    'vtag',
}
# Whether to suppress warnings related to unknown X- HTTP headers
_IGNORE_UNKNOWN_X_HEADERS = True

# When True, attempt to override JavaScript's Date class in HTML documents to
# return a consistent datetime for "now" matching the Date header that the
# page was served with.
# 
# Some pages like <https://newsletter.pragmaticengineer.com/archive> format
# URLs in JavaScript based on the current datetime. By pinning the current
# datetime to a consistent value, the generated URLs will also have a
# consistent value, allowing those generated URLs to be cached effectively.
_ENABLE_PIN_DATE_MITIGATION = True

# By default, browsers may cache Crystal-served assets for 1 hour
_CACHE_CONTROL_POLICY = 'max-age=3600'  # type: Optional[str]

_HTTP_COLON_SLASH_SLASH_RE = re.compile(r'(?i)https?://')


class _HttpServer(HTTPServer):
    project: Project
    verbosity: Verbosity
    stdout: TextIOBase | None
    
    @override
    def handle_error(self, request, client_address):
        # Print to stderr a message starting with
        # 'Exception occurred during processing of request from'
        return super().handle_error(request, client_address)


class _RequestHandler(BaseHTTPRequestHandler):
    # Prevent slow/broken request from blocking all other requests
    timeout = 1  # second
    
    server: _HttpServer
    
    @property
    def project(self) -> Project:
        server = self.server
        assert isinstance(server, _HttpServer)
        return server.project
    
    @property
    def _server_host(self) -> str:
        return 'localhost:%s' % self.server.server_port  # type: ignore[attr-defined]
    
    # === Request Properties ===
    
    @property
    def request_host(self) -> str:
        return self.headers.get('Host', self._server_host)
    
    @property
    def referer(self) -> str | None:
        return self.headers.get('Referer')
    
    # === Handle Incoming Request ===
    
    def parse_request(self):  # override
        # If we receive a request line with clearly binary data,
        # such as from an incoming HTTPS connection,
        # reject it immediately, before BaseHTTPRequestHandler
        # tries to print a ton of binary data to the console
        # as an error.
        if b'\x00' in self.raw_requestline:
            # Populate globals that send_error() requires
            self.command = None
            self.request_version = self.default_request_version
            self.close_connection = True
            self.requestline = '<binary data>'
            
            self.send_error(
                HTTPStatus.BAD_REQUEST,
                "Bad request version")
            return False
        
        return super().parse_request()
    
    def do_GET(self) -> None:  # override
        try:
            run_thread_switching_coroutine(
                SwitchToThread.BACKGROUND,
                self._do_GET())
        except BrokenPipeError:
            # Browser did drop connection before did finish sending response
            pass
    
    def _do_GET(self) -> Generator[SwitchToThread, None, None]:
        readonly = self.project.readonly  # cache
        dynamic_ok = (self.headers.get('X-Crystal-Dynamic', 'True') == 'True')
        
        # Parse self.path using RFC 2616 rules,
        # which in particular allows it to be an absolute URI!
        if self.path == '*':
            self.send_response(400)
            self.end_headers()
            return
        elif self.path.startswith('/'):
            host = self.headers.get('Host')
            if host is None:
                self.send_response(400)
                self.end_headers()
                return
        else:  # self.path must be an absolute URI
            pathurl_parts = urlparse(self.path)
            if pathurl_parts.scheme not in ('http', 'https') or pathurl_parts.netloc == '':
                self.send_response(400)
                self.end_headers()
                return
            
            # Rewrite self.path to be a URL path and the Host header to be a domain
            self.path = urlunparse(pathurl_parts._replace(scheme='', netloc=''))
            self.headers['Host'] = pathurl_parts.netloc  # replace any that existed before
        
        # Serve pin_date.js if requested
        m = _PIN_DATE_JS_PATH_RE.fullmatch(self.path)
        if m is not None:
            timestamp = int(m.group(1))
            
            self.send_response(200)
            self.send_header('Content-Type', 'application/javascript')
            self.end_headers()
            self.wfile.write(_pin_date_js(timestamp).encode('utf-8'))
            return
        
        # Serve resource revision in archive
        archive_url = self.get_archive_url(self.path)
        if archive_url is not None:
            yield SwitchToThread.FOREGROUND
            
            # If URL not in archive in its original form,
            # see whether it exists in the archive in a different form,
            # or whether it should be created in a different form
            resource = self.project.get_resource(archive_url)
            if resource is None:
                archive_url_alternatives = Resource.resource_url_alternatives(
                    self.project, archive_url)
                if len(archive_url_alternatives) >= 2:
                    assert archive_url_alternatives[0] == archive_url
                    # TODO: Optimize to use a bulk version of Project.get_resource()
                    #       rather than making several individual queries
                    for urla in archive_url_alternatives[1:]:
                        if self.project.get_resource(urla) is not None:
                            # Redirect to existing URL in archive
                            yield SwitchToThread.BACKGROUND
                            self.send_redirect(self.get_request_url(urla))
                            return
                    
                    # Redirect to canonical form of URL in archive
                    yield SwitchToThread.BACKGROUND
                    self.send_redirect(self.get_request_url(archive_url_alternatives[-1]))
                    return
            # (Either resource exists at archive_url, or archive_url is in canonical form)
            
            if resource is None:
                # If the previously undiscovered resource is a member of an
                # existing resource group, presume that the user is interested 
                # in downloading it immediately upon access
                matching_rg = self._find_group_matching_archive_url(archive_url)
                if matching_rg is not None and not readonly and dynamic_ok:
                    self._print_warning('*** Dynamically downloading new resource in group {!r}: {}'.format(
                        matching_rg.display_name,
                        archive_url,
                    ))
                    
                    # Try download resource immediately
                    def download_resource() -> Resource:
                        assert archive_url is not None
                        return Resource(self.project, archive_url)
                    resource = fg_call_and_wait(download_resource)
                    yield SwitchToThread.BACKGROUND
                    self._try_download_revision_dynamically(resource, needs_result=False)
                    # (continue to serve downloaded resource revision)
                else:
                    yield SwitchToThread.BACKGROUND
                    self.send_resource_not_in_archive(archive_url)
                    return
            
            if resource.definitely_has_no_revisions:
                revision = None  # type: Optional[ResourceRevision]
            else:
                def get_default_revision() -> ResourceRevision | None:
                    assert resource is not None
                    return resource.default_revision(
                        stale_ok=True if self.project.readonly else False
                    )
                revision = fg_call_and_wait(get_default_revision)
            
            yield SwitchToThread.BACKGROUND
            
            if revision is None:
                if not readonly and dynamic_ok:
                    # If the existing resource is also a root resource,
                    # presume that the user is interested 
                    # in downloading it immediately upon access
                    matching_rr = self._find_root_resource_matching_archive_url(archive_url)
                    if matching_rr is not None:
                        self._print_warning('*** Dynamically downloading root resource {!r}: {}'.format(
                            matching_rr.display_name,
                            archive_url,
                        ))
                    
                    # If the existing resource is a member of an
                    # existing resource group, presume that the user is interested 
                    # in downloading it immediately upon access
                    matching_rg = self._find_group_matching_archive_url(archive_url)
                    if matching_rg is not None:
                        self._print_warning('*** Dynamically downloading existing resource in group {!r}: {}'.format(
                            matching_rg.display_name,
                            archive_url,
                        ))
                    
                    if matching_rr is not None or matching_rg is not None:
                        # Try download resource immediately
                        revision = self._try_download_revision_dynamically(resource, needs_result=True)
                        # (continue to serve downloaded resource revision)
                
                if revision is None:
                    self.send_resource_not_in_archive(archive_url)
                    return
            
            # If client did make a conditional request which did match the revision,
            # send a short HTTP 304 response rather than the whole revision
            if_none_match = self.headers['If-None-Match']
            if if_none_match is not None:
                etag = revision.etag
                # NOTE: Only an If-None-Match containing a single ETag is supported for now
                if etag is not None and etag == if_none_match:
                    self.send_response(304)
                    self.end_headers()
                    return
            
            self.send_revision(revision, archive_url)
            return
        
        # Dynamically rewrite incoming link from archived resource revision
        referer = self.referer  # cache
        if referer is not None and dynamic_ok:
            referer_urlparts = urlparse(referer)
            if ((referer_urlparts.netloc == '' and referer_urlparts.path.startswith('/')) or 
                    referer_urlparts.netloc == self._server_host):
                referer_archive_url = self.get_archive_url(referer_urlparts.path)
                if referer_archive_url is not None:
                    referer_archive_urlparts = urlparse(referer_archive_url)
                    assert isinstance(referer_archive_urlparts.scheme, str)
                    assert isinstance(referer_archive_urlparts.netloc, str)
                    requested_archive_url = '{}://{}{}'.format(
                        referer_archive_urlparts.scheme,
                        referer_archive_urlparts.netloc,
                        self.path,
                    )
                    redirect_url = self.get_request_url(requested_archive_url)
                    
                    self._print_warning('*** Dynamically rewriting link from {}: {}'.format(
                        referer_archive_url,
                        requested_archive_url,
                    ))
                    
                    self.send_redirect(redirect_url, vary_referer=True)
                    return
        
        # Serve Welcome page
        path_parts = urlparse(self.path)
        if path_parts.path == '/':
            self.send_welcome_page(parse_qs(path_parts.query), vary_referer=True)
            return
        
        # Serve Not Found page
        self.send_not_found_page(vary_referer=True)
        return
    
    def _find_root_resource_matching_archive_url(self, archive_url: str) -> RootResource | None:
        for rr in self.project.root_resources:
            if rr.resource.url == archive_url:
                return rr
        return None
    
    def _find_group_matching_archive_url(self, archive_url: str) -> ResourceGroup | None:
        # ...excluding "do not download" groups
        for rg in self.project.resource_groups:
            if rg.contains_url(archive_url) and not rg.do_not_download:
                return rg
        return None
    
    @staticmethod
    @bg_affinity
    def _try_download_revision_dynamically(resource: Resource, *, needs_result: bool) -> ResourceRevision | None:
        try:
            return resource.download(
                # NOTE: Need to wait for embedded resources as well.
                #       If we were to serve a downloaded HTML page with
                #       embedded links that were not yet downloaded,
                #       they would be served as HTTP 404 until they
                #       finished downloading. To avoid serving a broken
                #       page we must wait longer for the embedded resources
                #       to finish downloading.
                wait_for_embedded=True,
                needs_result=needs_result,
                # Assume optimistically that resource is embedded so that
                # it downloads at "interactive priority", without inserting
                # any artificial delays
                is_embedded=True,
            ).result()
        except Exception:
            # Don't care if there was an error downloading
            return None
    
    # === Send Page ===
    
    @bg_affinity
    def send_welcome_page(self, query_params: dict[str, list[str]], *, vary_referer: bool) -> None:
        # TODO: Is this /?url=** path used anywhere anymore?
        if 'url' in query_params:
            archive_url = query_params['url'][0]
            redirect_url = self.get_request_url(archive_url)
            
            self.send_redirect(redirect_url, vary_referer=vary_referer)
            return
        
        self.send_response(200)
        self.send_header('Content-Type', 'text/html')
        if vary_referer:
            self.send_header('Vary', 'Referer')
        self.end_headers()
        
        self.wfile.write(dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
                <meta charset="utf-8" />
                <title>Welcome | Crystal</title>
            </head>
            <body>
                <p>Enter the URL of a page to load from the archive:</p>
                <form action="/">
                    URL: <input type="text" name="url" value="http://" /><input type="submit" value="Go" />
                </form>
            </body>
            </html>
            """
        ).lstrip('\n').encode('utf-8'))
    
    @bg_affinity
    def send_not_found_page(self, *, vary_referer: bool) -> None:
        self.send_response(404)
        self.send_header('Content-Type', 'text/html')
        if vary_referer:
            self.send_header('Vary', 'Referer')
        self.end_headers()
        
        self.wfile.write(dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
                <meta charset="utf-8" />
                <title>Not Found | Crystal</title>
            </head>
            <body>
                <p>There is no page here.</>
                <p>Return to <a href="/">home page</a>?</p>
            </body>
            </html>
            """
        ).lstrip('\n').encode('utf-8'))
    
    @bg_affinity
    def send_resource_not_in_archive(self, archive_url: str) -> None:
        self.send_response(404)
        self.send_header('Content-Type', 'text/html')
        self.end_headers()
        
        self.wfile.write((dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
                <meta charset="utf-8" />
                <title>Not in Archive | Crystal</title>
            </head>
            <body>
                <p>The requested resource was not found in the archive.</p>
                <p>The original resource is located here: <a href="%(archive_url)s">%(archive_url)s</a></p>
            </body>
            </html>
            """
        ).lstrip('\n') % {
            # TODO: Shouldn't this be HTML-escaped?
            'archive_url': archive_url
        }).encode('utf-8'))
        
        self._print_error('*** Requested resource not in archive: ' + archive_url)
    
    @bg_affinity
    def send_redirect(self, redirect_url: str, *, vary_referer: bool=False) -> None:
        self.send_response(307)
        self.send_header('Location', redirect_url)
        if vary_referer:
            self.send_header('Vary', 'Referer')
        self.end_headers()
    
    # === Send Revision ===
    
    @bg_affinity
    def send_revision(self, revision: ResourceRevision, archive_url: str) -> None:
        if revision.error is not None:
            self.send_resource_error(revision.error_dict, archive_url)
            return
        assert revision.has_body
        
        if revision.is_http:
            self.send_http_revision(revision)
        else:
            self.send_generic_revision(revision)
    
    @bg_affinity
    def send_resource_error(self, error_dict, archive_url) -> None:
        self.send_response(400)
        self.send_header('Content-Type', 'text/html')
        self.end_headers()
        
        self.wfile.write((dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
                <meta charset="utf-8" />
                <title>Fetch Error | Crystal</title>
            </head>
            <body>
                <p>
                    A <tt>%(error_type)s</tt> error with message <tt>%(error_message)s</tt>
                    was encountered when fetching this resource.
                </p>
                <p>The original resource is located here: <a href="%(archive_url)s">%(archive_url)s</a></p>
            </body>
            </html>
            """
        ).lstrip('\n') % {
            'error_type': (
                html_escape(error_dict['type'])
                if error_dict is not None
                else 'unknown'
            ),
            'error_message': (
                html_escape(error_dict['message'])
                if error_dict is not None and error_dict['message'] is not None
                else 'unknown'
            ),
            # TODO: Shouldn't this be HTML-escaped?
            'archive_url': archive_url
        }).encode('utf-8'))
        
        self._print_error('*** Requested resource was fetched with error: ' + archive_url)
    
    @bg_affinity
    def send_http_revision(self, revision: ResourceRevision) -> None:
        if revision.is_http_304:
            def do_resolve_http_304():
                nonlocal revision
                revision = revision.resolve_http_304()  # reinterpret
            fg_call_and_wait(do_resolve_http_304)
        
        metadata = revision.metadata
        assert metadata is not None
        
        # Determine Content-Type to send
        assert revision.has_body
        (doc, links, content_type_with_options) = revision.document_and_links()
        if not content_type_with_options:
            content_type_with_options = revision.declared_content_type_with_options
        
        # Determine headers to send
        headers = [[k, v] for (k, v) in metadata['headers']]  # clone, make mutable
        if content_type_with_options is not None:
            # Replace first preexisting 'Content-Type' header with new value,
            # or append new 'Content-Type' header to end of headers
            for kv in headers:
                (k, v) = kv
                if k.lower() == 'content-type':
                    kv[1] = content_type_with_options
                    break
            else:
                headers.append(['content-type', content_type_with_options])
        
        # Try extract revision datetime from Date header
        revision_datetime = revision.date
        
        # Send status line
        self.send_response_without_extra_headers(
            metadata['status_code'],
            metadata['reason_phrase'])
        
        # Send headers
        for (name, value) in headers:
            name_lower = name.lower()  # cache
            
            if name_lower == 'location':
                self.send_header(name, self.get_request_url(value))
                continue
                
            if name_lower in _HEADER_ALLOWLIST:
                self.send_header(name, value)
            else:
                if name_lower not in _HEADER_DENYLIST:
                    if not (_IGNORE_UNKNOWN_X_HEADERS and name_lower.startswith('x-')):
                        self._print_warning(
                            '*** Ignoring unknown header in archive: {}: {}'.format(name, value))
                continue
        if _CACHE_CONTROL_POLICY is not None:
            self.send_header('Cache-Control', _CACHE_CONTROL_POLICY)
        self.end_headers()
        
        # Send body
        self.send_revision_body(revision, doc, links, revision_datetime)
    
    @bg_affinity
    def send_generic_revision(self, revision: ResourceRevision) -> None:
        # Determine what Content-Type to send
        assert revision.has_body
        (doc, links, content_type_with_options) = revision.document_and_links()
        if not content_type_with_options:
            content_type_with_options = revision.declared_content_type_with_options
        
        # Send status line
        self.send_response(200)
        
        # Send headers
        if content_type_with_options is not None:
            self.send_header('Content-Type', content_type_with_options)
        self.end_headers()
        
        # Send body
        self.send_revision_body(revision, doc, links, revision_datetime=None)
    
    @bg_affinity
    def send_revision_body(self, 
            revision: ResourceRevision, 
            doc: Document | None, 
            links: list[Link], 
            revision_datetime: datetime.datetime | None
            ) -> None:
        # Send body
        if doc is None:
            # Not a document. Cannot rewrite content.
            with revision.open() as body:
                sock = self.connection
                try:
                    # NOTE: It would be more straightforward to use
                    #           shutil.copyfileobj(body, self.wfile)
                    #       but copyfileobj() does not use os.sendfile()
                    #       internally (which is the fastest file copy primitive):
                    #           https://github.com/python/cpython/issues/69249
                    sock.sendfile(body)
                except BrokenPipeError:
                    # Browser did disconnect early
                    return
        else:
            # Rewrite links in document
            base_url = revision.resource.url
            self._rewrite_links(links, base_url, self.get_request_url)
            
            if _ENABLE_PIN_DATE_MITIGATION:
                # TODO: Add try_insert_script() to Document interface
                if isinstance(doc, HtmlDocument) and revision_datetime is not None:
                    doc.try_insert_script(
                        _PIN_DATE_JS_PATH_PREFIX + 
                        str(int(revision_datetime.timestamp())))
            
            # Output altered document
            try:
                self.wfile.write(str(doc).encode('utf-8'))
            except BrokenPipeError:
                # Browser did disconnect early
                return
    
    @staticmethod
    def _rewrite_links(
            links: list[Link],
            base_url: str,
            get_request_url: Callable[[str], str]
            ) -> None:
        for link in links:
            relative_url = link.relative_url
            if relative_url.startswith('#'):
                # Don't rewrite links to anchors on the same page,
                # because some pages use JavaScript libraries
                # that treat such "local anchor links" specially
                pass
            else:
                if _HTTP_COLON_SLASH_SLASH_RE.fullmatch(relative_url):
                    absolute_url = relative_url
                else:
                    absolute_url = urljoin(base_url, relative_url)
                request_url = get_request_url(absolute_url)
                link.relative_url = request_url
    
    # === Send Response ===
    
    def send_response_without_extra_headers(self, code, message=None) -> None:
        """
        Similar to BaseHTTPRequestHandler.send_response(),
        but does not send its own {Server, Date} headers.
        
        The caller should still attempt to send its own Date header
        to conform with RFC 7231 §7.1.1.2, which requires that origin servers
        with a reliable clock to generate a Date header.
        """
        self.log_request(code)
        self.send_response_only(code, message)
    
    # === URL Transformation ===
    
    def get_archive_url(self, request_path: str) -> str | None:
        match = _REQUEST_PATH_IN_ARCHIVE_RE.match(request_path)
        if match:
            (scheme, rest) = match.groups()
            archive_url = '{}://{}'.format(scheme, rest)
            return archive_url
        else:
            # If valid default URL prefix is set, use it
            default_url_prefix = self.project.default_url_prefix  # cache
            if default_url_prefix is not None and not default_url_prefix.endswith('/'):
                assert request_path.startswith('/')
                return default_url_prefix + request_path
            
            return None
    
    def get_request_url(self, archive_url: str) -> str:
        return self.get_request_url_with_host(
            archive_url, self.request_host, self.project.default_url_prefix)
    
    @staticmethod
    def get_request_url_with_host(
            archive_url: str,
            request_host: str,
            default_url_prefix: str | None,
            ) -> str:
        """
        Given the absolute URL of a resource, returns the URL that should be used to
        request it from the archive server.
        """
        # If valid default URL prefix is set, use it
        if default_url_prefix is not None and not default_url_prefix.endswith('/'):
            if archive_url.startswith(default_url_prefix + '/'):
                return f'http://{request_host}' + archive_url[len(default_url_prefix):]
        
        archive_url_parts = urlparse(archive_url)
        
        request_scheme = 'http'
        request_netloc = request_host
        request_path = '_/{}/{}{}'.format(
            archive_url_parts.scheme,
            archive_url_parts.netloc, archive_url_parts.path)
        request_params = archive_url_parts.params
        request_query = archive_url_parts.query
        request_fragment = archive_url_parts.fragment
        
        request_url_parts = (
            request_scheme, request_netloc, request_path,
            request_params, request_query, request_fragment
        )
        request_url = urlunparse(request_url_parts)
        return request_url
    
    # === Logging ===
    
    def log_error(self, format, *args):  # override
        self._print_error(format % args)
    
    def log_message(self, format, *args):  # override
        self._print_info(format % args)
    
    # === Print ===
    
    def _print_info(self, message: str) -> None:
        if self._verbosity == 'indent':
            message = '    ' + message  # reinterpret
        print_info(message, file=self._stdout)
    
    def _print_warning(self, message: str) -> None:
        if self._verbosity == 'indent':
            message = '    ' + message  # reinterpret
        print_warning(message, file=self._stdout)
    
    def _print_error(self, message: str) -> None:
        if self._verbosity == 'indent':
            message = '    ' + message  # reinterpret
        print_error(message, file=self._stdout)
    
    @property
    def _verbosity(self) -> Verbosity:
        return self.server.verbosity
    
    @property
    def _stdout(self) -> TextIOBase | None:
        return self.server.stdout


_PIN_DATE_JS_TEMPLATE = dedent(
    """
    window.Date = (function() {
        const RealDate = window.Date;  // capture
        
        function PageLoadDate() {
            if (this === window) {
                // Date() -> str
                return (new PageLoadDate()).toString();
            } else {
                // new Date() -> Date
                return new RealDate(%d);
            }
        }
        PageLoadDate.now = function() {
            return (new PageLoadDate()).getTime();
        }
        PageLoadDate.UTC = RealDate.UTC
        
        return PageLoadDate;
    })();
    """
).lstrip()  # type: str

def _pin_date_js(timestamp: int) -> str:
    return _PIN_DATE_JS_TEMPLATE % timestamp


# ------------------------------------------------------------------------------


import os
import sys
from typing import List


class ListenableMixin:
    """
    Mixin for objects that have a listener list.
    """
    _WARN_IF_LEAKING_LISTENERS = \
        os.environ.get('CRYSTAL_LEAKING_LISTENER_WARNINGS', 'False') == 'True'
    
    # Optimize per-instance memory use, since some subclasses have very many objects
    __slots__ = (
        'listeners',
    )
    
    def __init__(self, *args, **kwargs) -> None:
        self.listeners = []  # type: List[object]
        super().__init__(*args, **kwargs)
    
    def __del__(self) -> None:
        if hasattr(self, '_WARN_IF_LEAKING_LISTENERS') and self._WARN_IF_LEAKING_LISTENERS:
            if hasattr(self, 'listeners') and self.listeners is not None:
                if len(self.listeners) != 0:
                    print(
                        f'*** Listenable object {self!r} still had listeners '
                            f'when it was finalized: {self.listeners!r}',
                        file=sys.stderr)
        if hasattr(super(), '__del__'):
            super().__del__()  # type: ignore[misc]



from typing import TYPE_CHECKING
import wx

if TYPE_CHECKING:
    import wx.adv


def fix_date_picker_size(dp: 'wx.adv.DatePickerCtrl') -> 'wx.adv.DatePickerCtrl':
    dp_best_size = dp.GetBestSize()
    dp.SetMinSize(wx.Size(
        # HACK: macOS, Windows: Add space to read last digit of year
        dp_best_size.width + 5,
        # HACK: macOS: Remove extra padding at top (which messes up centering)
        dp_best_size.height - 4))
    return dp



"""
Parses JSON documents.
"""

from __future__ import annotations

from crystal.doc.generic import Document, Link
from crystal.doc.html.soup import (
    ABSOLUTE_HTTP_LINK_RE, PROBABLE_EMBEDDED_URL_RE,
)
import json
from typing import BinaryIO


def parse_json_and_links(
        body_bytes: BinaryIO,
        declared_charset: str | None=None
        ) -> tuple[Document, list[Link]] | None:
    try:
        json_obj = json.load(body_bytes)  # type: ignore[attr-defined]
    except Exception:
        return None
    links = []  # type: list[Link]
    # TODO: If JSON document is just a str which is a link, recognize that link
    _find_links(json_obj, links)
    return (_JsonDocument(json_obj), links)


def _find_links(json_obj: object, links: list[Link]) -> None:
    if isinstance(json_obj, dict):
        # dict
        for (k, v) in json_obj.items():
            # Add links
            if isinstance(k, str) and ABSOLUTE_HTTP_LINK_RE.fullmatch(k) is not None:
                key_link = _DictKeyLink(json_obj, k)
                links.append(key_link)
            else:
                key_link = None
            if isinstance(v, str) and ABSOLUTE_HTTP_LINK_RE.fullmatch(v) is not None:
                value_link = _DictValueLink(json_obj, key_link or k)
                links.append(value_link)
            
            # Recurse
            _find_links(k, links)
            _find_links(v, links)
    
    elif isinstance(json_obj, list):
        # list
        for (i, item) in enumerate(json_obj):
            # Add links
            if isinstance(item, str) and ABSOLUTE_HTTP_LINK_RE.fullmatch(item) is not None:
                links.append(_ListItemLink(json_obj, i))
            
            # Recurse
            _find_links(item, links)
    
    else:
        # scalar
        pass


class _JsonDocument(Document):
    def __init__(self, json_obj) -> None:
        self._json_obj = json_obj
    
    def __str__(self) -> str:
        return json.dumps(self._json_obj)  # type: ignore[attr-defined]


class _JsonLink(Link):  # abstract
    def __init__(self, type_title: str) -> None:
        self.title = None
        self.type_title = type_title
        self.embedded = PROBABLE_EMBEDDED_URL_RE.search(self.relative_url) is not None


class _DictKeyLink(_JsonLink):
    def __init__(self, json_obj: dict, key: str) -> None:
        self._json_obj = json_obj
        self.key = key
        super().__init__('Dict Key Reference')
    
    def _get_relative_url(self) -> str:
        return self.key
    def _set_relative_url(self, url: str) -> None:
        old_key = self.key
        new_key = url
        value = self._json_obj[old_key]
        
        del self._json_obj[old_key]
        self._json_obj[new_key] = value
        
        self.key = new_key
    relative_url = property(_get_relative_url, _set_relative_url)


class _DictValueLink(_JsonLink):
    def __init__(self, json_obj: dict, key_ref: _DictKeyLink | str) -> None:
        self._json_obj = json_obj
        self._key_ref = key_ref
        super().__init__('Dict Value Reference')
    
    @property
    def _key(self) -> str:
        return (
            self._key_ref.key 
            if isinstance(self._key_ref, _DictKeyLink) 
            else self._key_ref
        )
    
    def _get_relative_url(self) -> str:
        return self._json_obj[self._key]
    def _set_relative_url(self, url: str) -> None:
        self._json_obj[self._key] = url
    relative_url = property(_get_relative_url, _set_relative_url)


class _ListItemLink(_JsonLink):
    def __init__(self, json_obj: list, index: int) -> None:
        self._json_obj = json_obj
        self._index = index
        super().__init__('List Item Reference')
    
    def _get_relative_url(self) -> str:
        return self._json_obj[self._index]
    def _set_relative_url(self, url: str) -> None:
        self._json_obj[self._index] = url
    relative_url = property(_get_relative_url, _set_relative_url)



from collections.abc import Callable, Iterator
from contextlib import contextmanager
from crystal.util import cli
import sys
import traceback
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import wx


def bind(
        window: 'wx.Window', 
        event_type, 
        target: 'Callable[[wx.Event], None]',
        *args
        ) -> None:
    """
    Equivalent to wx.Window.Bind(), but safer.
    """
    window.Bind(event_type, _bind_target()(target), *args)


@contextmanager
def _bind_target() -> Iterator[None]:
    """
    Decorates any function that is a target of wx.Window.Bind().
    
    Will handle any uncaught exceptions by printing their traceback
    and swallowing them to prevent them from bubbling up to wx.
    
    It has been observed at least on macOS that uncaught exceptions
    that make it to wx can put it into an invalid state that will
    cause a segmentation fault later (especially when Python exits).
    """
    try:
        yield
    except BaseException:
        err_file = sys.stderr
        print(cli.TERMINAL_FG_RED, end='', file=err_file)
        print('Exception in wxPython listener:', file=err_file)
        traceback.print_exc(file=err_file)
        print(cli.TERMINAL_RESET, end='', file=err_file)
        err_file.flush()

